import string
import regex as re
import nldk.load_nltk

from nltk.corpus import stopwords
from nltk.stem import SnowballStemmer, WordNetLemmatizer
from nltk.tokenize import word_tokenize

content_cleaner_default_language = "english"


class NLPContentCleaner(object):
    def __init__(
        self,
        default_language=None,
        tokenize=None,
        remove_html=None,
        remove_punctuation=None,
        remove_extra_space=None,
        convert_lowercase=None,
        remove_stopwords=None,
        stem=None,
        lemmatize=None,
        lemmatize_pos=None,
    ):
        """
        Constructs a ContentCleaner instance

        @param tokenize (Boolean): Tells the clean() function whether to return the cleaned input as tokens
        or as a whole sentence concatenated again. Defaults to False.

        """
        # Get lists of particular language availability
        self._stopwords_languages = list(stopwords.fileids())
        self._stemmer_languages = list(SnowballStemmer.languages)

        # Until we build a package, we will be limited to lemmatizing in English only
        self._lemmatizer = WordNetLemmatizer()
        self._lemmatizer_languages = ["english"]

        #
        # A Note on the WordNetLemmatizer:
        #
        # Lemmatizer lets you select part of speech!
        # This is really cool, because it means you pick the part of speech to devolve your words into.
        # However, because the lemmatizer currently only supports English, this feature isn't as
        # useful as it would be.
        #
        # This lemmatizer is state of the art (circa 2023), but it is not definitive yet. More work needs to be
        # done in lemmatization to improve data preparation.
        #
        # Example:
        # self._lemmatizer.lematize(word, pos)
        # word (str) – The input word to lemmatize.
        # pos (str) – The Part Of Speech tag. Valid options are “n” for nouns, “v” for verbs, “a” for adjectives,
        # “r” for adverbs and “s” for satellite adjectives.
        #

        # Unless specified, assign default values to options
        self._default_language = (
            default_language if default_language else content_cleaner_default_language
        )
        self._return_tokenized = tokenize if tokenize is not None else False
        self._remove_html = remove_html if remove_html is not None else False
        self._remove_punctuation = (
            remove_punctuation if remove_punctuation is not None else False
        )
        self._remove_extra_space = (
            remove_extra_space if remove_extra_space is not None else False
        )
        self._convert_lowercase = (
            convert_lowercase if convert_lowercase is not None else False
        )
        self._remove_stopwords = (
            remove_stopwords if remove_stopwords is not None else False
        )
        self._stem = stem if stem is not None else False
        self._lemmatize = lemmatize if lemmatize is not None else False

        # Defaults to lemmatizing with the default lemmatizer part of speech
        self._lemmatize_pos = lemmatize_pos if lemmatize_pos is not None else None

    def get_lemmatize_pos_supported(self):
        return (
            {"pos": "n", "meaning": "nouns"},
            {"pos": "v", "meaning": "verbs"},
            {"pos": "a", "meaning": "adjectives"},
            {"pos": "r", "meaning": "adverbs"},
            {"pos": "s", "meaning": "satellite adjectives"},
        )

    def get_languages_supported(self):
        """
        @returns (List[str]): the list of languages that will be supported by the features indicated
            by the parameters of the class instance.
        """
        all_languages = (
            self._stopwords_languages
            + self._stemmer_languages
            + self._lemmatizer_languages
        )
        return list(
            set(self._stopwords_languages if self._remove_stopwords else all_languages)
            & set(self._stemmer_languages if self._stem else all_languages)
            & set(self._lemmatizer_languages if self._lemmatize else all_languages)
        )

    def clean(self, content, language=None, verbose_output=None):
        """
        A series of hand-picked cleanup operations to achieve the following basic results:
        1. Removing HTML characters, ASCII
        2. Remove extra punctuation
        3. Remove extra space characters
        4. Convert to lowercase
        5. Remove stop words
        6. Stem (SnowballStemmer)
        7. Lemmatize (WordNetLemmatizer, only supported for language="english")

        @param content (str): a standard Python unicode string to be cleaned.
        @param language (str): the language that the string is known to be in.
        @param verbose_output(str): display verbose output at each step of cleaning.

        For a list of supported languages, use the get_languages_supported() instance method.
        """

        language = language if language else self._default_language
        verbose_output = verbose_output if verbose_output is not None else False

        if not language:
            raise ValueError(
                "Missing required parameter 'language'. Must be supplied at instantiation time or when calling clean(...)."
            )

        if verbose_output:
            print(f"Content prior to cleaning: {content}")

        # Remove HTML (this is sufficient, but naive. TODO: find more sophisticated / accurate HTML removal)
        if self._remove_html:
            content = re.sub("<.*?>", " ", content)
            if verbose_output:
                print(f"Content after removing html: {content}")

        # Remove all punctuation (kind of a hack of str.maketrans which requires the first two positional
        # parameters, but also adds functionality to remove anything of the type passed in the third, which
        # in this case is string.punctuation)
        if self._remove_punctuation:
            content = content.translate(str.maketrans(" ", " ", string.punctuation))
            if verbose_output:
                print(f"Content after removing punctuation: {content}")

        # Remove any space character including newline with a single space
        if self._remove_extra_space:
            content = re.sub(r"\s+", r" ", content)
            if verbose_output:
                print(f"Content after removing extra space: {content}")

        # Convert to lowercase
        if self._convert_lowercase:
            content = content.lower()
            if verbose_output:
                print(f"Content after converting to lowercase: {content}")

        # Tokenize content
        content = word_tokenize(content)

        if verbose_output:
            print(f"Content after tokenizing: {content}")

        language = language.lower()

        # If language specified and in allowed list of languages, remove stop words (otherwise skip)
        if self._remove_stopwords and language in self._stopwords_languages:
            stop_words = set(stopwords.words(language))
            content = [word for word in content if word not in stop_words]
            if verbose_output:
                print(f"Content after removing stopwords: {content}")

        # Stem content
        if self._stem and language in self._stemmer_languages:
            stemmer = SnowballStemmer(language)
            content = [stemmer.stem(word) for word in content]
            if verbose_output:
                print(f"Content after stemming: {content}")

        # Lemmatize content (supports only English at this time)
        # TODO: write or find package for multilingual lemmatization (have not found one yet)
        if self._lemmatize and language in self._lemmatizer_languages:
            if self._lemmatize_pos is not None:
                content = [
                    self._lemmatizer.lemmatize(word, self._lemmatize_pos)
                    for word in content
                ]
            else:
                content = [self._lemmatizer.lemmatize(word) for word in content]
            if verbose_output:
                print(f"Content after lemmatization: {content}")

        return content if self._return_tokenized else " ".join(content)


def get_lemmatize_pos_supported():
    cleaner = NLPContentCleaner()
    return cleaner.get_lemmatize_pos_supported()


def content_cleaner_languages_supported(**kwargs):
    """
    Returns the languages supported by the supplied feature set, using NLPContentCleaner class
    defaults when parameter not supplied.

    NOTE: Only these three options have any effect on language suppport, which is why the other
    initialization options have been left out.

    @param **kwargs:
        See below

    Keyword Arguments:
        'remove_stopwords' (Boolean): Include stopword removal in the list of features for which supported
        languages are returned.
        'stem' (Boolean): Include stemming in the list of features for which supported languages are returned.
        'lemmatize' (Boolean): Include lemmatization in the list of features for which supported languages
        are returned.
        NOTE: All other arguments of NLPContentCleaner are accepted, but have no effect on language support.

    @returns (List[str]): the list of languages that will be supported by the features indicated
        by the boolean parameters above. I"f none are supplied, it is assumed a list of languages
        supported by all features is desired.
    """
    remove_stopwords = kwargs.get("remove_stopwords", None)
    stem = kwargs.get("stem", None)
    lemmatize = kwargs.get("lemmatize", None)

    cleaner = NLPContentCleaner(
        remove_stopwords=remove_stopwords, stem=stem, lemmatize=lemmatize
    )

    return cleaner.get_languages_supported()


def clean_content_file(
    in_path,
    out_path,
    language=None,
    tokenize=None,
    remove_html=None,
    remove_punctuation=None,
    remove_extra_space=None,
    convert_lowercase=None,
    remove_stopwords=None,
    stem=None,
    lemmatize=None,
    lemmatize_pos=None,
    verbose_output=None,
):
    cleaner = NLPContentCleaner(
        default_language=language,
        tokenize=tokenize,
        remove_html=remove_html,
        remove_punctuation=remove_punctuation,
        remove_extra_space=remove_extra_space,
        convert_lowercase=convert_lowercase,
        remove_stopwords=remove_stopwords,
        stem=stem,
        lemmatize=lemmatize,
        lemmatize_pos=lemmatize_pos,
    )

    with open(in_path, "r") as input_file:
        with open(out_path, "w") as output_file:
            for line in input_file:
                output_file.write(
                    cleaner.clean(content=line, verbose_output=verbose_output)
                )
    print(
        f"Finished cleaning '{language if language else content_cleaner_default_language}' language input file '{in_path}'. Clean output written to file '{out_path}'"
    )

    """
    Th above call to cleaner.clean executes a series of hand-picked cleanup operations
    to achieve the following basic results:
        1. Removing HTML characters, ASCII
        2. Remove extra punctuation
        3. Remove extra space characters
        4. Convert to lowercase
        5. Remove stop words
        6. Stem (SnowballStemmer)
        7. Lemmatize (WordNetLemmatizer, only supported for language="english")

    @param content (str): a standard Python unicode string to be cleaned.
    @param language (str): the language that the string is known to be in.
    @param verbose_output(str): display verbose output at each step of cleaning.

    For a list of supported languages, use the get_languages_supported() function.
    """



# All known languages according to Wikipedia as of 2023
all_languages = [
    "afrikaans",
    "albanian",
    "amharic",
    "arabic",
    "armenian",
    "assamese",
    "azerbaijani",
    "basque",
    "belarusian",
    "bengali",
    "bhojpuri",
    "bosnian",
    "brazilian-portuguese",
    "bulgarian",
    "burmese",
    "cantonese",
    "catalan",
    "cebuano",
    "chechen",
    "chinese",
    "croatian",
    "czech",
    "danish",
    "dari",
    "dinka",
    "dutch",
    "english",
    "estonian",
    "finnish",
    "flemish",
    "french",
    "fula",
    "galician",
    "georgian",
    "german",
    "greek",
    "gujarati",
    "haitian",
    "hausa",
    "hebrew",
    "hiligaynon",
    "hindi",
    "hmong",
    "hungarian",
    "icelandic",
    "igbo",
    "ilokano",
    "indonesian",
    "irish",
    "italian",
    "japanese",
    "javanese",
    "kannada",
    "kazakh",
    "khmer",
    "kirundi",
    "korean",
    "kurdish",
    "kyrgyz",
    "lao",
    "latin",
    "latvian",
    "lithuanian",
    "luxembourgish",
    "macedonian",
    "madurese",
    "maithili",
    "malagasy",
    "malay",
    "malayalam",
    "maltese",
    "mandarin",
    "marathi",
    "mongolian",
    "nepalese",
    "norwegian",
    "odia",
    "oromo",
    "pangasinense",
    "pashto",
    "persian",
    "polish",
    "portuguese",
    "punjabi",
    "quechua",
    "romanian",
    "russian",
    "sardinian",
    "serbian",
    "serbo-croatian",
    "sindhi",
    "sinhala",
    "slovak",
    "slovene",
    "somali",
    "spanish",
    "sundanese",
    "swahili",
    "swedish",
    "tagalog",
    "tahitian",
    "taiwanese",
    "tamil",
    "tatar",
    "telugu",
    "thai",
    "tibetan",
    "turkish",
    "ukrainian",
    "urdu",
    "uyghur",
    "uzbek",
    "vietnamese",
    "welsh",
    "wolof",
    "wu",
    "xhosa",
    "yiddish",
    "yoruba",
    "zulu",
]



from nldk.nlp_content_cleaner import NLPContentCleaner
from nldk.tmx import parse_tmx_file
from nldk.nlp_content_cleaner import content_cleaner_languages_supported
from nldk.all_languages import all_languages
from nldk.write_csv_line import write_csv_line


def tmx2csv(
    in_path,
    out_path,
    languages,
    clean_languages=None,
    tmx_verbose_output=None,
    cleaner_verbose_output=None,
    **kwargs,
):
    """
    Converts a multilingual (minimum 1 language) TMX translation file to CSV. Optionally, the output data can be cleaned
    using `NLPContentCleaner`. All of the options available for instantiation of the `NLPContentCleaner` class are
    therefore included in the parameter list for this function.

    @params in_filepath (str): The filepath to the input TMX file
    @params out_filepath (str): The filepath to the output CSV file
    @params languages (str): Space-separated list of languages included in TMX file, full name (ie "english" or "japanese")
    """
    # Remove 'verbose_output' - require user to send `cleaner_verbose_output` explicitly
    if kwargs.get("verbose_output", None):
        del kwargs["verbose_output"]

    languages = languages.split()

    clean_idx2lang = {}
    if clean_languages:
        clean_languages = clean_languages.split()
        for language in clean_languages:
            if not language in languages:
                raise ValueError(
                    f"Language '{language}' specified in `--clean_languages` was not one of the TMX file languages specified by "
                    + "`--languages`. Can only clean language columns that exist in the source TMX file."
                )
            clean_idx2lang[languages.index(language)] = language
        cleaner = NLPContentCleaner(**kwargs)

    print(
        f"Starting conversion{' and cleaning' if clean_languages else ''} of TMX file '{in_path}' to CSV file '{out_path}'..."
    )

    with open(out_path, "w") as csv_file:

        def write_raw_items(items):
            write_csv_line(csv_file, items)

        def write_clean_items(items):
            clean_items = []
            for idx, item in enumerate(items):
                clean_lang = clean_idx2lang.get(idx, None)
                if clean_lang:
                    clean_items.append(
                        cleaner.clean(
                            item, clean_lang, verbose_output=cleaner_verbose_output
                        )
                    )
                else:
                    clean_items.append(item)
            write_csv_line(csv_file, clean_items)

        parse_tmx_file(
            in_path,
            column_names=languages,
            verbose_output=tmx_verbose_output,
            write_row=write_clean_items if clean_languages else write_raw_items,
        )


def tmx2csv_lang_support(**kwargs):
    """
    Given the same parameters that can be passed to tmx2csv(), return an array of language names
    that are supported by the selected features.

    """
    if kwargs.get("clean_languages", None):
        return content_cleaner_languages_supported(**kwargs)
    else:
        return all_languages



import nltk

# Any NLTK packages that need to be specifically downloaded go here. NLTK will
# generally throw a readable error letting you know if something needs to be
# added to this list.
nltk.download("stopwords")
nltk.download("wordnet")



def get_basic_object():
    """
    Literally just get an object that you can use like `object.some_key = 3`.
    This should not be this obtuse. JavaScript is far superior with things
    like this.
    """
    return type("basic_object", (object,), {})()



def write_csv_line(file_object, items):
    """
    Given an open file object and list of strings, add them as a properly-formatted and
    escaped line to the CSV file object.

    @param file_object (io.TextIOWrapper): a target file object to write the line of CSV data
    @param items (List[str]): a list of strings to be added in order as a line to the CSV file
    @returns csv_line: the actual csv_line (including newline) that was written to the CSV file
    """

    # Format the line to write
    csv_line = (
        # Escape any double quotes as double double quotes
        ",".join(['"' + value.replace('"', '""') + '"' for value in items])
        + "\n"
    )

    file_object.write(csv_line)

    # Return formatted line back to caller
    return csv_line



from xml.sax import parse, SAXParseException
from xml.sax.handler import ContentHandler
import io, os


class TMXColumnParser(ContentHandler):
    def __init__(
        self,
        xml_file_object,
        file_size_in_bytes,
        column_names=None,
        write_row=None,
        verbose_output=None,
        status_update_frequency_bytes=None,
        content_direct_parent=None,
        content_row_divider=None,
        experimental=None,
    ):
        """

        xml_file_object                     File object (already opened) that the parser will be working off of

        file_size_in_bytes                  File size in bytes of the file to which xml_file_object refers

        write_row                           Lambda to perform write of individual row of TMX data (one translation unit
                                            of semantically equivalent segments differing by language) Default is to print the
                                            list of extracted content values to the console.

                                            Note: if a column does not have a value, an empty string will be placed in that
                                            ordinal position in the list of content values received.

                                            Note: columns will appear in the same order as they are given in the constructor
                                            parameter `column_names`

        verbose_output                   Boolean to determine whether to output verbosely to console

        status_update_frequency_bytes       Number of bytes to traverse before updating the status bar with a new percentage.
                                            Defaults to file_size_in_bytes / 1000 for updates every tenth of a percent. Set
                                            to 0 to disable status updates specifically.

        content_direct_parent               Tells the parser which element directly contains the sections of content.
                                            Defaults to "seg" for TMX files.

        content_row_divider                 Tells the parser which element represents the division between rows of content.
                                            Defaults to "tu" for TMX files (as various translations of one sentence are often
                                            contained within one enclosing <tu>...</tu> tag).

        column_names                        List of column names. Informs the parser of:
                                                - How many languages (or columns, generally) are contained within each row.
                                                - What to call those columns if that data becomes useful in future enhancements

        experimental                        Boolean value to toggle experimental functionality on and off. Default is False.

        """

        FILE_SIZE_MINIMUM_BYTES = 10

        self._experimental = experimental if experimental else False

        # Validate that xml_file_object is indeed an instance of io.TextIOWrapper class
        if not isinstance(xml_file_object, io.TextIOWrapper):
            raise ValueError(
                "Parameter 'xml_file_object' must be member of io.TextIOWrapper class (must be a file object)"
            )

        # Validate that file_size_in_bytes has been provided
        if not file_size_in_bytes:
            raise ValueError("Parameter 'file_size_in_bytes [long]' is required")

        file_size_in_bytes = int(file_size_in_bytes)

        # Validate that file_size_in_bytes is coercible to a positive number greater than
        # or equal to FILE_SIZE_MINIMUM_BYTES
        if file_size_in_bytes < FILE_SIZE_MINIMUM_BYTES:
            raise ValueError(
                f"Parameter 'file_size_in_bytes [long]' must be at least {FILE_SIZE_MINIMUM_BYTES}"
            )

        # Validate that column_names has been provided and is a list
        if not column_names:
            raise ValueError("Parameter 'column_names [list of strings]' is required")
        if not type(column_names) == list:
            raise ValueError("Parameter 'column_names' must be a list of strings")

        self._verbose_output = verbose_output if verbose_output else False
        self._xml_file_object = xml_file_object
        self._file_size_in_bytes = file_size_in_bytes

        self._status_update_frequency_bytes = (
            status_update_frequency_bytes
            if status_update_frequency_bytes
            else file_size_in_bytes / 1000
        )

        self._content_direct_parent = (
            content_direct_parent if content_direct_parent else "seg"
        )

        self._content_row_divider = content_row_divider if content_row_divider else "tu"
        self._content_column_names = column_names
        self._content_column_names_len = len(self._content_column_names)
        self._write_row = (
            write_row
            if write_row
            else lambda values, column_names: print(values, column_names)
        )

        # Store iteratively each current piece of content (to be written when content_row_divider is closed)
        self._current_values = []

        if self._experimental:
            # Currently not used, experimental only. An array built by creating a new current_element_parents
            # beginning with the currently-starting element, and followed by all previously currently-starting
            # elements in antichronological order.
            self._current_element_parents = []

        # Signal to the characters() method that content should be saved into self._current_values
        # ultimately to be written when the end of the current content_row_divider element is encountered
        self._is_content = False

        # Used internally to keep track of when another status update should happen (adds the value supplied
        # in self._status_update_frequency_bytes each time the status is updated and awaits that number of
        # bytes or more for the next iterative update)
        self._next_status_update_threshold = 0

        if self._verbose_output:
            print(vars(self))

        super(TMXColumnParser, self).__init__()

    def startDocument(self):
        if self._verbose_output:
            print(f"Starting parse on '{self._xml_file_object.name}'...")
        super(TMXColumnParser, self).startDocument()

    def endDocument(self):
        if self._verbose_output:
            print(f"Parse finished on on '{self._xml_file_object.name}'...")
        super(TMXColumnParser, self).endDocument()

    # Event fired when an XML element starts
    def startElement(self, name, attrs):
        if self._verbose_output:
            print(f"opening element: {name}")

        self._starting_element = name

        if self._experimental:
            # Experimental code only. self._current_element_parents is not used, but may be used
            # in a future enhancement to deal with deep hierarchies
            self._current_element_parents.insert(
                0, {"name": self._starting_element, "attrs": attrs}
            )

        if self._starting_element == self._content_direct_parent:
            # Always ensure self._current_values[current_idx] defaults to an empty string
            self._current_values.append("")

            # Let characters() method know that any characters accepted until further notice
            # should be considered content
            self._is_content = True

        super(TMXColumnParser, self).startElement(name, attrs)

    # Event fired when an XML element ends
    def endElement(self, name):
        if self._verbose_output:
            print(f"closing element: {name}")

        if self._experimental:
            # Experimental only (not used):
            # Remove the most recent entry from the hierarchy of elements
            test_closing_element = self._current_element_parents.pop(0)

            # Experimental only (not used):
            if not closing_element == test_closing_element["name"]:
                raise ValueError(
                    f"Expected closing element '{test_closing_element}' popped from current_element_parents stack to equal actual closing element '{closing_element}'"
                )

        closing_element = name

        # If a content direct parent element is closing, signal to the characters() method that viable content
        # is not currently available
        if closing_element == self._content_direct_parent:
            self._is_content = False

        # If a content divider element is closing, write the content element using the supplied write_row lambda
        # and set stack of current values back to an empty array
        if closing_element == self._content_row_divider:
            self._write_row(self._current_values)
            self._current_values = []

        """
        
        Show status bar every time self._status_update_frequency_bytes number of bytes are eclipsed

        """
        current_file_position_bytes = self._xml_file_object.tell()

        if (
            self._status_update_frequency_bytes > 0
            and current_file_position_bytes > self._next_status_update_threshold
        ):
            self._next_status_update_threshold += self._status_update_frequency_bytes

            # parameter 'end' causes it to print to the same line if not being interrupted by other console logs
            print(
                f"Status: {current_file_position_bytes / self._file_size_in_bytes * 100:.2f}%\r",
                end="",
            )

        super(TMXColumnParser, self).endElement(name)

    # Event fired once for every element regardless of content
    def characters(self, content):
        if self._is_content and self._starting_element == self._content_direct_parent:
            # This is a potentially desired piece of content.

            # We should only adding content to the last element in the array
            current_idx = len(self._current_values) - 1

            if self._verbose_output:
                print(f"content of current element: {content}")
                print(f"current_idx: {current_idx}")

            # Check if current column index is out of bounds of the self._content_column_names array
            if current_idx >= self._content_column_names_len:
                raise ValueError(
                    f"Input TMX file is malformed or your parameters to your instance of {type(self).__name__ } are not representative of the data elements contained in your TMX file."
                )

            # Concatenate any new content with the existing self._current_values[...]
            if content:
                self._current_values[current_idx] += content

        super(TMXColumnParser, self).characters(content)


"""
parse_tmx_xml_file_object

Functional wrapper for TMXColumnParser. Uses defaults of TMXColumnParser
unless overriden

"""


def parse_tmx_xml_file_object(
    xml_file_object,
    file_size_in_bytes,
    verbose_output=None,
    status_update_frequency_bytes=None,
    write_row=None,
    column_names=None,
    content_direct_parent=None,
    content_row_divider=None,
    experimental=None,
):
    content_handler = TMXColumnParser(
        xml_file_object,
        file_size_in_bytes,
        verbose_output=verbose_output,
        status_update_frequency_bytes=status_update_frequency_bytes,
        column_names=column_names,
        write_row=write_row,
        content_direct_parent=content_direct_parent,
        content_row_divider=content_row_divider,
        experimental=experimental,
    )
    try:
        parse(xml_file_object, content_handler)
    except SAXParseException as err:
        print("Error during SAX parsing:", err)


"""
LIMITATION: As of right now, this only works for TMX files when specifying ALL of the column_names
in the file. To get a select range of column indices, or even better a select list of language
abbreviations, some enhancements will need to be made, but this works to get a basic TMX file
in its entirety over to CSV or perform some other iterative parsing action.

Parses any TMX file according to a lambda that defines what needs to happen when a row (dict)
is written (either to CSV file, to console, or to in-memory structure like dataframe).

"""


def parse_tmx_file(
    tmx_filepath,
    column_names,
    write_row=None,
    verbose_output=None,
    # Override these to customize how the TMX (XML) file is parsed if wanting to use
    # the parser in a different way
    status_update_frequency_bytes=None,
    content_direct_parent=None,
    content_row_divider=None,
):
    with open(tmx_filepath) as file:
        parse_tmx_xml_file_object(
            file,
            file_size_in_bytes=os.stat(tmx_filepath).st_size,
            column_names=column_names,
            write_row=write_row,
            verbose_output=verbose_output,
            status_update_frequency_bytes=status_update_frequency_bytes,
            content_direct_parent=content_direct_parent,
            content_row_divider=content_row_divider,
        )



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name="nldk",
    version="0.1.0-alpha.1",
    description="Tools for handling baseline tasks related to natural language processing.",
    long_description="Tools for handling baseline tasks related to natural language processing.",
    url="https://github.com/vanguardapps/nldk",
    download_url="https://pypi.python.org/pypi/nldk",
    license="MIT",
    author="Roy McClanahan",
    author_email="rmcclanahan@vanguardapps.io",
    maintainer="Roy McClanahan",
    maintainer_email="rmcclanahan@vanguardapps.io",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Utilities",
        "Topic :: Text Processing :: General",
    ],
    packages=find_packages(),
    install_requires=[
        "regex==2023.10.3",
        "nltk==3.8.1",
    ],
    scripts=["nldk/bin/nldk"],
)



def get_file_lines(filepath, start_line, end_line):
    """
    Gets a range of lines from a text file located at filepath, returns as array of strings (lines)

    """
    line_count = 0
    lines = []
    with open(filepath, "r") as file:
        for line in file:
            if line_count >= start_line and line_count <= end_line:
                lines.append(line)
            elif line_count > end_line:
                break
            line_count += 1
    return lines


def print_file_lines(in_path, start_line, end_line, out_path):
    """
    Sends a select range of lines in a file to `stdout` as-is.

    @param in_path (str):           The path to the file.
    @param start_line (int):        The beginning of the line range (included in output)
    @param end_line (int):          The end of the line range (not included in output)
    @param out_path (str):          The desired output filepath.

    """
    start_line = int(start_line)
    end_line = int(end_line)
    with open(out_path, "w") as output_file:
        output_file.writelines(get_file_lines(in_path, start_line, end_line))
    print(
        f"Output '{in_path}' lines {start_line} through {end_line} to new file '{out_path}'"
    )



from nldk.write_csv_line import write_csv_line


def tsv2csv(in_path, out_path):
    """
    Converts TSV files to CSV files.

    @param in_path (str):   The filepath to the input TSV file.
    @param out_path (str):  The filepath to the desired output CSV file.

    LIMITATION: will not handle CRLF characters in the TSV, as it reads file line by line.
    This is not important for most datasets, and does not negatively affect results using
    the data included in this repo.

    """
    with open(out_path, "w") as output_file:
        with open(in_path) as input_file:
            for line in input_file:
                write_csv_line(output_file, line.rstrip().split("\t"))
    print(f"Converted '{in_path}' from TSV to CSV. Result stored in '{out_path}'")



import argparse
import regex as re
from nldk.basic_object import get_basic_object


class CLIMenu(object):
    def __init__(self):
        # Disable default `-h`/`--help` behavior for `argparse.ArgumentParser`
        self._parser = argparse.ArgumentParser(add_help=False)
        self._options = {}
        self._all_parameters = set()

        # Having disabled `--help` default beahvior, we must define our own `--help` argument now.
        self._all_parameters.add("help")
        self._parser.add_argument(f"--help", action="store_true")

        # Support the `--languages-supported` flag by default
        self._all_parameters.add("languages_supported")
        self._parser.add_argument(f"--languages_supported", action="store_true")

    def add_primary_flag(
        self,
        flag=None,
        help=None,
        callback=None,
        get_languages_supported=None,
        required_params=None,
        optional_params=None,
    ):
        """
        Adds an option `--<flag>` to the `CommandLineMenu` instance. If the option is encountered, all of
        the parameters from `required_params` and `optional_params` will be concatenated and passed from the
        user's input to `callback`. `required_params` will be validated to ensure their presence (more specific
        validations must be performed inside of `callback`)

        @param flag (str):                          Indicates the primary option flag as a string. For instance, if the primary
                                                    option is `--convert_to_csv`, you would pass `flag="convert_to_csv"`.
                                                    NOTE: Any non-alphanumeric character in flag will be coerced to an underscore.

        @param help (str):                          The string text to be displayed next to the primary utility flag when `--help` is
                                                    passed to the application.

        @param callback (function):                 The callback function that will be invoked if the user passes `--<flag>` when
                                                    executing the script. Using the `--convert_to_csv` example, this may be a callback
                                                    `convert_to_csv(in_filename, out_filename)` (note all hyphens `-` will be converted
                                                    to underscores `_` for the purpose of variable names passed to callbacks). Callback
                                                    is responsible for accepting all arguments listed in the union of `required_params`
                                                    and `optional_params` as keyword arguments. NOTE: All parameters will have to be cast
                                                    to desired type from `str`.

        @param get_languages_supported (function):  A callback function called with all required and optional params of the chosen main
                                                    utility that must return the list of languages (lowercase full name) that support
                                                    all of the specified required and optional parameters. If the chosen main utility has
                                                    any default values for parameters, those should be used accordingly to make the
                                                    determination of whether a language is supported or not. The goal is to make it so
                                                    invoking the `--languages_supported` flag will show the exact list of languages supported
                                                    by the feature flags if the same flags were used to actually invoke the utility.

        @param required_params (Dict):              A dict where the keys are the names of the required parameters and the values are the
                                                    help text to be displayed next to that parameter when user passes `--help`. The string
                                                    `(bool)` may be appended to the end of any key (parameter name) to indicate that it is
                                                    a true/false flag as opposed to a parameter that expects a value. For instance
                                                    `is_collated(bool)` would specify a parameter `--is_collated` that is intended to be
                                                    used as a true/false flag. Leaving `(bool)` off implies that the parameter will be used
                                                    to capture a value. Required params are always validated to ensure they are present.
                                                    NOTE: Any non-alphanumeric character in required_params will be coerced to an underscore.

        @param optional_params (Dict):              A dict where the keys are the names of the optional parameters and the values are the
                                                    help text to be displayed next to that parameter when user passes `--help`. The string
                                                    `(bool)` may be appended to the end of any key (parameter name) to indicate that it is
                                                    a true/false flag as opposed to a parameter that expects a value. For instance
                                                    `is_collated(bool)` would specify a parameter `--is_collated` that is intended to be
                                                    used as a true/false flag. Leaving `(bool)` off implies that the parameter will be used
                                                    to capture a value. NOTE: Any non-alphanumeric character in any key of `required_params`
                                                    will be coerced to an underscore `_`.

        """

        if flag is None:
            raise ValueError("Missing required parameter `flag`")

        if flag in self._all_parameters:
            raise ValueError(
                f"Flag `--{flag}` has already been added as an option. Cannot add the same option twice."
            )

        if callback is None:
            raise ValueError("Missing required parameter `callback`")

        if get_languages_supported is None:
            raise ValueError("Missing required parameter `get_languages_supported`")

        if not (required_params or optional_params):
            raise ValueError(
                "Must specify one or both: `required_params` and/or `optional_params`"
            )

        # Sanitize primary flag to only allow alphanumeric characters and underscores
        flag = re.sub("[^0-9a-zA-Z]+", "_", flag)

        # Provide a default help text
        help = help if help else f"TODO: write `help` text for the `--{flag}` flag"

        # Add this primary utility flag to the parser
        self._parser.add_argument(f"--{flag}", action="store_true", help=help)

        if required_params:
            # 1. Strip required params
            required_params = {
                key.strip(): value for key, value in required_params.items()
            }
            # 2. Acknowledge required params intended to be boolean flags
            required_param_is_boolean = {
                key: True for key in required_params.keys() if key.endswith("(bool)")
            }
            # 3. Remove "(bool)" out if it appears
            required_params = {
                key.replace("(bool)", ""): value
                for key, value in required_params.items()
            }
            # 4. Sanitize required parameter names to only allow alphanumeric characters and underscores
            required_params = {
                re.sub("[^0-9a-zA-Z]+", "_", key): value
                for key, value in required_params.items()
            }
            # Add all required params as either boolean or value-capturing
            for param_name in required_params.keys():
                if param_name not in self._all_parameters:
                    self._all_parameters.add(param_name)
                    self._parser.add_argument(
                        f"--{param_name}",
                        action="store_true"
                        if required_param_is_boolean.get(param_name, None)
                        else None,
                    )

        if optional_params:
            # 1. Strip optional params
            optional_params = {
                key.strip(): value for key, value in optional_params.items()
            }

            # 2. Acknowledge optional params intended to be boolean flags
            optional_param_is_boolean = {
                key.replace("(bool)", ""): True
                for key in optional_params.keys()
                if key.endswith("(bool)")
            }

            # 3. Remove "(bool)" out if it appears
            optional_params = {
                key.replace("(bool)", ""): value
                for key, value in optional_params.items()
            }

            # 4. Sanitize optional parameter names to only allow alphanumeric characters and underscores
            optional_params = {
                re.sub("[^0-9a-zA-Z]+", "_", key): value
                for key, value in optional_params.items()
            }

            # Add all optional params as either boolean or value-capturing
            for param_name in optional_params.keys():
                if param_name not in self._all_parameters:
                    self._all_parameters.add(param_name)
                    self._parser.add_argument(
                        f"--{param_name}",
                        action="store_true"
                        if optional_param_is_boolean.get(param_name, None)
                        else None,
                    )

        if required_params and optional_params:
            # Validate that there is no overlap between `required_params` and `optional_params`
            overlapping_params = set(required_params.keys()) & set(
                optional_params.keys()
            )
            if overlapping_params:
                raise ValueError(
                    f"Parameters [`--{'`, `--'.join(overlapping_params)}] are inclued in both `required_params` and `optional_params`. "
                    + "There should be no overlap between `required_params` and `optional_params`."
                )

        # Add the parameter attributes to our dict of options
        self._options[flag] = get_basic_object()  # Creates a basic object
        self._options[flag].callback = callback
        self._options[flag].required_params = required_params
        self._options[flag].optional_params = optional_params
        self._options[flag].help = help
        self._options[flag].get_languages_supported = get_languages_supported

    def execute(self):
        if not self._options:
            raise ValueError(
                "Cannot execute `CommandLineMenu` because no options have been added"
            )

        args = self._parser.parse_args()

        primary_flags = self._options.keys()
        user_selection = None

        for flag in primary_flags:
            if getattr(args, flag, None):
                if user_selection:
                    raise ValueError("Only one primary flag may be selected at a time!")
                user_selection = flag

        help_flag = True if getattr(args, "help") else False

        if not user_selection:
            if help_flag:
                # Display general help output for all primary utility flags
                output = (
                    "\nUse one of the following primary flags. Use a primary flag in combination"
                    + " with `--help` to view help information specific to that flag:\n"
                )
                output += (
                    "--help: Display help for either the whole application (without a primary option) or for the "
                    + "primary option indicated by the feature flag (ie `--clean_content_file`).\n"
                )
                output += (
                    "--languages_supported: Display a list of languages that are supported for the features indicated for the "
                    + "primary utility selected (list of strings, lowercase, full language names).\n"
                )
                for flag, option in self._options.items():
                    output += f"--{flag}: {option.help}\n"
                print(output + "\n")
                return
            else:
                # If no `--help` flag was sent, throw an error as a selection from primary flags is required
                raise ValueError(
                    f"At least one primary utility must be passed. Should be one of [`--help`, `--{'`, `--'.join(primary_flags)}]"
                )

        callback = self._options[user_selection].callback
        required_params = self._options[user_selection].required_params
        optional_params = self._options[user_selection].optional_params
        get_languages_supported = self._options[user_selection].get_languages_supported
        user_selection_help = self._options[user_selection].help

        if help_flag:
            # Display help related to the selected primary flag
            output = f"\nHelp for `--{user_selection}`: {user_selection_help}\n\n"

            if required_params:
                output += "Required parameters:\n"
                for flag, help in required_params.items():
                    output += f"--{flag}: {help}\n"
            if optional_params:
                output += "\nOptional parameters:\n"
                for flag, help in optional_params.items():
                    output += f"--{flag}: {help}\n"
            print(output)
            return

        # Only using parameter keys from here onward, so converting to keys-only lists
        required_params = required_params.keys() if required_params else None
        optional_params = optional_params.keys() if optional_params else None

        # Dict to hold all parameters for the action being executed
        params = {}

        missing_required_params = []
        if required_params:
            # Store required parameters (check for missing)
            for name in required_params:
                value = getattr(args, name, None)
                if value:
                    params[name] = value
                else:
                    missing_required_params.append(name)

        if optional_params:
            # Store optional parameters
            for name in optional_params:
                value = getattr(args, name, None)
                if value:
                    params[name] = value

        # Check to see if the `--languages-supported` was specified
        if getattr(args, "languages_supported", None):
            language_support_output = f"\nLanguages supported for `--{user_selection}` using your parameters:\n\n"
            languages_list = get_languages_supported(**params)
            if languages_list:
                language_support_output += f"{languages_list}"
            else:
                language_support_output += "No languages are suppported for the specified utility and parameters."
            print(language_support_output + "\n")
            return
        else:
            # Check for missing required parameters
            if len(missing_required_params) > 0:
                raise ValueError(
                    f"Missing required parameters `--{'`, `--'.join(missing_required_params)}"
                )

            # Execute callback with params dict converted to keyword arguments
            callback(**params)
